/*
 * steve diverdi
 * cs 125
 * project 3
 *
 * common values for project 3
 */

#ifndef COMMON_HH
#define COMMON_HH


// INCLUDES-------------------------------------------------------------


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>				// for file io control
#include <errno.h>				// error handling

#include <iostream.h>				// standard io
#include <strings.h>				// string functions

#include <sys/types.h>
#include <sys/time.h>				// for the timeval struct
#include <sys/stat.h>
#include <sys/socket.h>				// socket functions
#include <sys/select.h>				// select command

#include <netinet/in.h>                         // internet protocol
#include <netdb.h>                              // for hostent


#include "sockets.hh"				// my socket functions


// DEFINES -------------------------------------------------------------


#define SERVER_PORT     20141			// my uid for server port
#define MAX_PACKET	516			// max packet size
#define MAX_DATA	MAX_PACKET - 4		// data is 4 less than  packet
#define RETRIES		5			// retries after timeout

// packet types
#define RRQ	1
#define WRQ	2
#define DATA	3
#define ACK	4
#define ERROR	5

// transfer modes
#define OCTET	 10
#define NETASCII 11
#define MAIL	 12

// error codes
#define TNOTDEF	 0
#define TNOFILE	 1
#define TACCESS	 2
#define TDSKFUL	 3
#define TILGLOP	 4
#define TBADTID	 5
#define TEXISTS	 6
#define TNOUSER	 7


// CONSTANTS -----------------------------------------------------------


// the duration to use for the time out value
static timeval TIMEOUT = { 5, 0 };


// STRUCTS -------------------------------------------------------------


// basic packet structure - a 2 bytes opcode followed by 510 bytes of data
struct packet
{
	uint16_t	opcode;
	char		data[MAX_PACKET - 2];
};


// rrq and wrq packets have no extra fields
typedef packet rrq_packet;
typedef packet wrq_packet;


// data packet has an additional 2 byte block number
struct data_packet 
{
	uint16_t	opcode;
	uint16_t	block;
	char		data[MAX_DATA];
};


// ack packet is only a two byte opcode and a two byte block number
struct ack_packet 
{
	uint16_t	opcode;
	uint16_t	block;
};


// error packet has a two byte errorcode after the opcode
struct error_packet 
{
	uint16_t	opcode;
	uint16_t	errorcode;
	char		errmsg[MAX_PACKET - 4];
};


// FUNCTIONS -----------------------------------------------------------

// send an error packet, display error on stderr
	void		error(int, char*, int, struct sockaddr*, int);

// network byte order and host byte order functions
	void		networkify(void*, int);
	void		denetworkify(void*, int);

// packet forming functions
	void		makeRRQ(struct packet*, int*, char*, char*);
	void		makeWRQ(struct packet*, int*, char*, char*);
	void		makeDATA(struct data_packet*, int*, int, char*, int);
	void		makeACK(struct ack_packet*, int*, int);
	void		makeERROR(struct error_packet*, int*, int, char*);

// get fields from packets
	int		getType(struct packet*);
	int		getBlock(struct packet*);
	char*		getFilename(struct packet*);
	int		getMode(struct packet*);
	char*		getErrormsg(struct packet*);

// socket ready time out
	bool		timeout(int);

// send and receive acknowledgement packets
	int		sendACK(int, int, struct sockaddr*, int, bool);
	int		recvACK(int, int, int, struct sockaddr*, int*);

// send and receive data packets
        int             recvDATA(int, struct data_packet*, int*, int, int,
                                 struct sockaddr*, int*);
        int             sendDATA(int, int, char*, int, struct sockaddr*, int);


#endif // COMMON_HH
