/*
 * steve diverdi
 * cs125 project 4
 *
 * tftp server
 * see RFC 1350
 */


// INCLUDES ------------------------------------------------------------

#include "common.hh"

// ---------------------------------------------------------------------

	int		main(int, char**);
	int		serve(int, struct packet*, int, struct 
                              sockaddr*, int);

	int		serveRRQ(int, char*, struct sockaddr*, int);
	int		serveWRQ(int, char*, struct sockaddr*, int);

// ---------------------------------------------------------------------

/*
 * Top level call. Creates a socket and starts listening for request 
 * packets.  When one is received, it forks a new process and hands the
 * request off to it to service.  Returns 0 on success, -1 on failure.
 */
int main (
    int         argc,
    char*       argv[])
{
    // create server socket
    int s = createSocket(SERVER_PORT);

    // make room for the incoming packet
    struct packet p;
    int len;
    struct sockaddr from;
    int fromlen = sizeof(from);

    // so long as there's more incoming packets, keep receiving 'em.
    while ((len = recvfrom(s, &p, sizeof(p), 0, &from, &fromlen)) >= 0)
    {
        // create a new process and give it the packet to deal with
        if (fork() == 0)
        {
            serve(s, &p, len, &from, fromlen);
            exit(0);
        }
    }

    return 0;
}


// ---------------------------------------------------------------------

/*
 * Parses the request packet and hands off to a read or write server
 * function.  Takes as arguments a pointer to the request packet, the
 * length of the request packet, the address of the client and the 
 * length of the client's address
 */
int serve (
    int			s,
    struct packet*	_p,
    int			_len,
    struct sockaddr*	client_addr,
    int			client_len)
{
    // put the packet in host byte order
    denetworkify(_p, _len);

    // create a new socket to handle the transfer on
    s = createSocket(SERVER_PORT + 1);

    // check the type of the packet
    if ((getType(_p) == RRQ) ||
        (getType(_p) == WRQ))
    {
        // this server only supports octet mode
        if (getMode(_p) != OCTET)
            cerr << "octet mode only" << endl;

        // get the filename from the packet
        char* filename = getFilename(_p);

        // hand off to a function to handle the read or write request 
        // specifically
        if (getType(_p) == RRQ)
            return serveRRQ(s, filename, client_addr, client_len);
        else
            return serveWRQ(s, filename, client_addr, client_len);
    }
    else
    {
        // not a request packet, so error
        error(TNOTDEF, "Illegal request", s, client_addr, client_len);
        return -1;
    }
}


/*
 * Serves a read request - reads a file from the server and sends it
 * to the client.  Takes as arguments a socket id, the filename to send,
 * the client's address, and the address length.  Returns 0 on success,
 * -1 on failure.
 */
int serveRRQ (
    int			s,
    char*		filename,
    struct sockaddr*	client_addr,
    int			client_len)
{
    cout << "Read " << filename << " initiated" << endl;

    // open the local file to read
    int file;
    if ((file = open(filename, O_RDONLY)) == -1)
    {
        error(TNOTDEF, "Couldn't open file for reading", s,
              client_addr, client_len);

        return -1;
    }

    int client_tid = ((struct sockaddr_in*) client_addr)->sin_port;
    int blockNum = 1;
    char blockData[MAX_DATA];
    int blen = MAX_DATA;

    // continue to loop while we're reading in maximum length blocks of 
    // data.  we know we're done when we get a block of data less than
    // the maximum length
    while (blen == MAX_DATA)
    {
        // read the next block of data and send it to the client
        blen = read(file, blockData, MAX_DATA);
        if (sendDATA(s, blockNum, blockData, blen, client_addr, client_len) 
                     == -1)
            return -1;

        // receive the client's acknowledgement
        if (recvACK(s, blockNum, client_tid, client_addr, &client_len) == -1)
            return -1;

        blockNum++;
    }

    // all done, so close the file and return 0
    cout << "Read " << filename << " successful" << endl;
    close(file);
    return 0;
}


/*
 * Serves a write request - takes a file from the client and writes it
 * to a local file.  Takes as arguments a socket id, the filename to
 * write, the client's address, and the length of the address.  Returns
 * 0 on success, -1 on failure.
 */
int serveWRQ (
    int			s,
    char*		filename,
    struct sockaddr*	client_addr,
    int			client_len)
{
    cout << "Write " << filename << " initiated" << endl;

    // open the local file for writing
    int file;
    mode_t mode = S_IRUSR|S_IWUSR;
    if ((file = creat(filename, mode)) == -1)
    {
        if (errno == EACCES)
            error(TACCESS, "Access violation", s, client_addr, client_len);
        else if (errno == EDQUOT)
            error(TDSKFUL, "Disk full or allocation exceeded", s,
                  client_addr, client_len);
        else
            error(TNOTDEF, "Couldn't open file for writing", s,
                  client_addr, client_len);

        return -1;
    }

    int client_tid = ((struct sockaddr_in*) client_addr)->sin_port;
    int blockNum = 0;
    int blen;
    struct data_packet incoming;
    int ilen = MAX_PACKET;
 
    // continue to receive packets as long as they're the maximum length.
    // we know we're done reading when we receive a packet that's less
    // than the full possible number of bytes.
    while (ilen == MAX_PACKET)
    {                
        // send an acknowledgement for the last packet received
        if (sendACK(s, blockNum, client_addr, client_len, true) == -1)
            return -1;

        // waiting on the next packet now...
        blockNum++;

        // receive the next data packet
        if (recvDATA(s, &incoming, &ilen, blockNum, client_tid,
                     client_addr, &client_len) == -1)
            return -1;
        
        // write the data to our file
        if ((blen = write(file, incoming.data, ilen - 4)) == -1)
        {
            if (errno == EDQUOT)
                error(TDSKFUL, "Disk full or allocation exceeded", s,
                      client_addr, client_len);
            else
                error(TNOTDEF, "Couldn't write file", s, client_addr,
                      client_len);
    
            return -1;
        }
    }

    // send our final acknowledgement
    if (sendACK(s, blockNum, client_addr, client_len, false) == -1)
        return -1;

    // all done, so close the file and return 0
    cout << "Write " << filename << " successful" << endl;
    close(file);
    return 0;
}

